/*
	GENERAL SGML PARSER CODE

	(c) COPYRIGHT MIT 1995 as part of the W3O reference library
	Please first read the full copyright statement in the file COPYRIGHT.
  
	This module implements an HTStream object. To parse an
	SGML file, create this object which is a parser. The object
	is (currently) created by being passed a DTD structure,
	and a target HTStructured object at which to throw the parsed stuff.

	adjusted for use in ProWesS reader
	by Joachim Van der Auwera
	for PROGS, Professional & Graphical Software
	started 14 November 1995
	30.11.1995 fixed a problem cfr newline in SGML_LITERAL between tags
	11.11.1998 entities with numbers between 160 and 255 now handled properly
*/

#include "mem_h"
#include "str_h"
#include "ctype_h"

/* Library include files */
#include "HTChunk_h"
#include "SGML_h"

#define INVALID (-1)
#define catch(x) if (err=(x)) return err; else

/* The State (context) of the parser */
/* This is passed with each call to make the parser reentrant */

	
/*		Element Stack
**		-------------
**	This allows us to return down the stack reselecting styles.
**	As we return, attribute values will be garbage in general.
*/
typedef struct _HTElement HTElement;
struct _HTElement {
	HTElement * next;	/* Previously nested element or 0 */
	HTTag * 	tag;	/* The tag at this level  */
};


typedef enum _sgml_state {
	S_text, S_literal, S_nl_literal, S_tag, S_tag_gap,
	S_attr, S_attr_gap, S_equals, S_value, S_after_open,
	S_nl, S_nl_tago,
	S_ero, S_cro,
	S_squoted, S_dquoted, S_end, S_entity, S_junk_tag
} sgml_state;


/*	Internal Context Data Structure */
/*	------------------------------- */

struct _HTStream {
	const HTStreamClass *isa;		/* inherited from HTStream */
	
	const SGML_dtd *	dtd;
	HTStructuredClass * actions;	/* target class  */
	HTStructured *		target; 	/* target object */

	HTTag * 			current_tag;
	int 				current_attribute_number;
	HTChunk *			string;
	HTElement * 		element_stack;
	sgml_state			state;
	char present[MAX_ATTRIBUTES];	/* Flags: attribute is present? */
	char * value[MAX_ATTRIBUTES];	/* allocated strings or NULL if none */
};


#define PUTC(ch) ((*context->actions->put_character)(context->target, ch))


/*	Find Attribute Number */
/*	--------------------- */

int SGMLFindAttribute (HTTag *tag, const char *s)
{
	attr *attributes = tag->attributes;
	int high, low, i, diff; 	/* Binary search for attribute name */

	for ( low=0, high=tag->number_of_attributes ;
		  high > low ;
		  diff < 0 ? (low = i+1) : (high = i) )
	{
		i = (low + (high-low)/2);
		STRCompareCI(attributes[i].name, s, &diff);
		if (diff==0) return i;		/* success: found it */
	}
	
	return -1;
}


/*	Handle Attribute */
/*	---------------- */

static void handle_attribute_name(HTStream *context, const char *s)
{
	HTTag * tag = context->current_tag;
	int i = SGMLFindAttribute(tag, s);

	if (i>=0)
	{
		context->current_attribute_number = i;
		context->present[i] = TRUE;
		MEMRelease(context->value[i]);
		context->value[i] = NULL;
		return;
	}
	context->current_attribute_number = INVALID;	/* Invalid */
}


/*	Handle attribute value */
/*	---------------------- */

static Error handle_attribute_value(HTStream *context, const char *s)
{
	Error err=ERR_OK;

	if (context->current_attribute_number != INVALID)
	{
		Size len=strlen(s);
		char **dest=context->value+context->current_attribute_number;

		err=MEMAllocate(len,dest);
		if (err) return err;
		STRCopy(s,*dest);
	}
	/* there can only be one assignment */
	context->current_attribute_number = INVALID;
	return err;
}

/* list of the entities for numbers 160-255, maps to entity name */
static const char *entities[]={
	"nbsp",
	"iexcl",
	"cent",
	"pound",
	"curren",
	"yen",
	"brvbar",
	"sect",
	"uml",
	"copy",
	"ordf",
	"laquo",
	"not",
	"shy",
	"reg",
	"macr",
	"deg",
	"plusmn",
	"sup2",
	"sup3",
	"acute",
	"micro",
	"para",
	"middot",
	"cedil",
	"sup1",
	"ordm",
	"raquo",
	"frac14",
	"frac12",
	"frac34",
	"iquest",
	"Agrave",
	"Aacute",
	"Acirc",
	"Atilde",
	"Auml",
	"Aring",
	"AElig",
	"Ccedil",
	"Egrave",
	"Eacute",
	"Ecirc",
	"Euml",
	"Igrave",
	"Iacute",
	"Icirc",
	"Iuml",
	"ETH",
	"Ntilde",
	"Ograve",
	"Oacute",
	"Ocirc",
	"Otilde",
	"Ouml",
	"times",
	"Oslash",
	"Ugrave",
	"Uacute",
	"Ucirc",
	"Uuml",
	"Yacute",
	"THORN",
	"szlig",
	"agrave",
	"aacute",
	"acirc",
	"atilde",
	"auml",
	"aring",
	"aelig",
	"ccedil",
	"egrave",
	"eacute",
	"ecirc",
	"euml",
	"igrave",
	"iacute",
	"icirc",
	"iuml",
	"eth",
	"ntilde",
	"ograve",
	"oacute",
	"ocirc",
	"otilde",
	"ouml",
	"divide",
	"oslash",
	"ugrave",
	"uacute",
	"ucirc",
	"uuml",
	"yacute",
	"thorn",
	"yuml"
};
static void handle_entity_name(HTStream *context, const char *s)
{
	const char **entities = context->dtd->entity_names;
	int high, low, i, diff;

	for ( low=0, high = context->dtd->number_of_entities ;
		  high > low ;
		  diff < 0 ? (low = i+1) : (high = i) )
	{
		/* Binary search */
		i = (low + (high-low)/2);
		STRCompareCD(entities[i], s, &diff);
		if (diff==0)
		{
			/* success: found it */
			(*context->actions->put_entity)(context->target, i);
			return;
		}
	}
	/* If entity string not found, display as text */
	PUTC('&');
	{
		const char *p;
		for (p=s; *p; p++)
		{
			PUTC(*p);
		}
	}
}


/*	Handle entity
**	-------------
**
** On entry,
**	s	contains the entity name zero terminated
** Bugs:
**	If the entity name is unknown, the terminator is treated as
**	a printable non-special character in all cases, even if it is '<'
*/
static void handle_entity(HTStream *context, char term)
{
	const char **entities = context->dtd->entity_names;
	const char *s = context->string->data;
	int high, low, i, diff;

	for ( low=0, high = context->dtd->number_of_entities ;
		  high > low ;
		  diff < 0 ? (low = i+1) : (high = i) )
	{
		/* Binary search */
		i = (low + (high-low)/2);
		STRCompareCD(entities[i], s, &diff);
		if (diff==0)
		{
			/* success: found it */
			(*context->actions->put_entity)(context->target, i);
			return;
		}
	}
	/* If entity string not found, display as text */
	PUTC('&');
	{
		const char *p;
		for (p=s; *p; p++)
		{
			PUTC(*p);
		}
	}
	PUTC(term);
}


/*	End element */
/*	----------- */

static void end_element(HTStream *context, HTTag *old_tag)
{
	if (old_tag->contents == SGML_EMPTY)
		return;

	while (context->element_stack)
	{
		/* Loop is error path only */
		HTElement *N = context->element_stack;
		HTTag *t = N->tag;
	
		if (old_tag != t)
		{
			/* Mismatch: syntax error */
			if (!context->element_stack->next)
				return;
		}

		context->element_stack = N->next;		/* Remove from stack */
		MEMRelease(N);
		(*context->actions->end_element)(context->target, t-context->dtd->tags);
		if (old_tag == t) return;	/* Correct sequence */
	
		/* Syntax error path only */
	}
}


/*	Start an element */
/*	---------------- */

static Error start_element(HTStream *context)
{
	Error err;
	HTTag * new_tag = context->current_tag;
	
	catch( (*context->actions->start_element)(
			context->target,
			new_tag - context->dtd->tags,
			context->present,
			context->value ) );
	if (new_tag->contents != SGML_EMPTY)
	{
		/* i.e. tag not empty */
		HTElement *N;
		catch( MEMAllocate(sizeof(HTElement),&N) );
		N->next = context->element_stack;
		N->tag = new_tag;
		context->element_stack = N;
	}
	return ERR_OK;
}


/*		Find Tag in DTD tag list
**		------------------------
**
** On entry,
**	dtd points to dtd structire including valid tag list
**	string	points to name of tag in question
**
** On exit,
**	returns:
**		NULL		tag not found
**		else		address of tag structure in dtd
*/
HTTag *SGMLFindTag(const SGML_dtd *dtd, const char *string)
{
	int high, low, i, diff;
	for( low=0, high=dtd->number_of_tags ;
		 high > low ;
		 diff < 0 ? (low = i+1) : (high = i) )
	{
		/* Binary search */
		i = (low + (high-low)/2);
		STRCompareCI(dtd->tags[i].name, string, &diff);
		if (diff==0)
		{
			/* success: found it */
			return &dtd->tags[i];
		}
	}
	return NULL;
}

/*________________________________________________________________________
**			Public Methods
*/


/*	Could check that we are back to bottom of stack! @@  */
Error SGML_flush(HTStream *context)
{
	while (context->element_stack)
	{
		HTElement *ptr = context->element_stack;

		context->element_stack = ptr->next;
		MEMRelease(ptr);
	}
	return (*context->actions->flush)(context->target);
}

Error SGML_free(HTStream * context)
{
	Error err;
	int cnt;

	while (context->element_stack)
	{
		/* Make sure, that all tags are gone */
		HTElement *ptr = context->element_stack;

		context->element_stack = ptr->next;
		MEMRelease(ptr);
	}
	if (err=(*context->actions->_free)(context->target))
		return err;
	HTChunkFree(context->string);
	for(cnt=0; cnt<MAX_ATTRIBUTES; cnt++)		 /* Leak fix Henrik 18/02-94 */
		if(context->value[cnt])
			MEMRelease(context->value[cnt]);
	MEMRelease(context);
	return err;
}

Error SGML_abort(HTStream *context, Error e)
{
	int cnt;
	while (context->element_stack)
	{
		/* Make sure, that all tags are gone */
		HTElement *ptr = context->element_stack;

		context->element_stack = ptr->next;
		MEMRelease(ptr);
	}
	(*context->actions->abort)(context->target, e);
	HTChunkFree(context->string);
	for(cnt=0; cnt<MAX_ATTRIBUTES; cnt++)		/* Leak fix Henrik 18/02-94 */
		if(context->value[cnt])
			MEMRelease(context->value[cnt]);
	MEMRelease(context);
	return e;
}


/*	Read and write user callback handle
**	-----------------------------------
**
**	 The callbacks from the SGML parser have an SGML context parameter.
**	 These calls allow the caller to associate his own context with a
**	 particular SGML context.
*/

Error SGML_character(HTStream *context, char c)
{
	Error err=ERR_OK;
	const SGML_dtd *dtd = context->dtd;
	HTChunk *string = context->string;

	switch(context->state) {
	
	case S_after_open:
		/* Strip one trailing newline */
		/* only after opening nonempty element.  - SGML:Ugh! */

		if (c=='\n' && (context->current_tag->contents != SGML_EMPTY))
			break;
		context->state = S_text;
		goto normal_text;
	
	case S_text:
normal_text:
		if (c=='&' && (!context->element_stack || (
				 context->element_stack->tag  &&
				 ( context->element_stack->tag->contents == SGML_MIXED ||
				   context->element_stack->tag->contents == SGML_RCDATA ) )))
		{
			string->size = 0;
			context->state = S_ero;
		}
		else
		if (c=='<')
		{
			string->size = 0;
			context->state =
				( context->element_stack && context->element_stack->tag  &&
				context->element_stack->tag->contents == SGML_LITERAL ) ?
					S_literal : S_tag;
		}
		else
		if (c=='\n')
		{
			/* Newline - ignore if before tag end! */
			context->state = S_nl;
		}
		else
			PUTC(c);
		break;

	case S_nl:
		if (c=='<')
		{
			string->size = 0;
			context->state =
				( context->element_stack && context->element_stack->tag  &&
				context->element_stack->tag->contents == SGML_LITERAL ) ?
					S_nl_literal : S_nl_tago;
		}
		else
		{
			PUTC('\n');
			context->state = S_text;
			goto normal_text;
		}
		break;

	case S_nl_tago:
		/* Had newline and tag opener */
		if (c != '/')
		{
			PUTC('\n'); 	/* Only ignore newline before </ */
		}
		context->state = S_tag;
		goto handle_S_tag;

	/*	In literal mode, waits only for specific end tag! */
	/*	Only for compatibility with old servers. */

	case S_literal :
	case S_nl_literal :
		catch( HTChunkPutc(string, c) );
		if ( toupper(c) != ( (string->size==1) ?
				'/' : context->element_stack->tag->name[string->size-2] ) )
		{
			int i;
		
			/*	If complete match, end literal */
			if ((c=='>') && (!context->element_stack->tag->name[string->size-2]))
			{
				end_element(context, context->element_stack->tag);
				string->size = 0;
				context->current_attribute_number = INVALID;
				context->state = S_text;
				break;
			}
			/* If Mismatch: recover string. */
			if (context->state==S_nl_literal) PUTC('\n');
			PUTC('<');
			for (i=0; i<string->size; i++)
			   PUTC(string->data[i]);
			context->state = S_text;
		}
		break;

	/*	Character reference or Entity */

   case S_ero:
		if (c=='#')
		{
			context->state = S_cro; 	/* &# is Char Ref Open */
			break;
		}
		context->state = S_entity;		/* Fall through! */
	
	/*	Handle Entities */

	case S_entity:
		if (isalnum(c))
			catch( HTChunkPutc(string, c) );
		else
		{
			catch( HTChunkTerminate(string) );
			handle_entity(context, c);
			context->state = S_text;
		}
		break;

	/*	Character reference */

	case S_cro:
		if (isalnum(c))
			catch( HTChunkPutc(string, c) ); /* accumulate a character code */
		else
		{
			int value, count;
			catch( HTChunkTerminate(string) );
			STRFormatGet(string->data, &count, "%d", &value);
			if (count==1)
			{
				if (value>=160 && value<=255)
				{
					handle_entity_name(context, entities[value-160]);
				}
				else
					PUTC(value);
			}
			context->state = S_text;
		}
		break;

	/* Tag */

	case S_tag: 					/* new tag */
handle_S_tag:
		if (isalnum(c))
			catch( HTChunkPutc(string, c) );
		else
		{
			/* End of tag name */
			HTTag * t;
			if (c=='/')
			{
				context->state = S_end;
				break;
			}
			catch( HTChunkTerminate(string) );

			t = SGMLFindTag(dtd, string->data);
			if (!t)
			{
				context->state = (c=='>') ? S_text : S_junk_tag;
				break;
			}
			context->current_tag = t;
			
			/* Clear out attributes */
			{
				int i;
				for (i=0; i< context->current_tag->number_of_attributes; i++)
					context->present[i] = FALSE;
			}
			string->size = 0;
			context->current_attribute_number = INVALID;
			
			if (c=='>')
			{
				if (context->current_tag->name)
					catch( start_element(context) );
				context->state = S_after_open;
			}
			else
			{
				context->state = S_tag_gap;
			}
		}
		break;

	case S_tag_gap:
		/* Expecting attribute or > */
		if (isspace(c))
		{
			/* Gap between attributes */
			break;
		}
		if (c=='>')
		{
			/* End of tag */
			if (context->current_tag->name)
				catch( start_element(context) );
			context->state = S_after_open;
			break;
		}
		catch( HTChunkPutc(string, c) );
		context->state = S_attr;		/* Get attribute */
		break;
	
	/* accumulating value */

	case S_attr:
		if (isspace(c) || (c=='>') || (c=='='))
		{
			/* End of word */
			catch( HTChunkTerminate(string) );
			handle_attribute_name(context, string->data);
			string->size = 0;
			if (c=='>')
			{
				/* End of tag */
				if (context->current_tag->name)
					catch( start_element(context) );
				context->state = S_after_open;
				break;
			}
			context->state = (c=='=' ?	S_equals: S_attr_gap);
		}
		else
		{
			catch( HTChunkPutc(string, c) );
		}
		break;
		
	case S_attr_gap:
		/* Expecting attribute or = or > */
		if (isspace(c))
		{
			/* Gap after attribute */
			break;
		}
		if (c=='>')
		{
			/* End of tag */
			if (context->current_tag->name)
				catch( start_element(context) );
			context->state = S_after_open;
			break;
		}
		else
		if (c=='=')
		{
			context->state = S_equals;
			break;
		}
		catch( HTChunkPutc(string, c) );
		context->state = S_attr;		/* Get next attribute */
		break;
	
	case S_equals:
		/* After attr = */
		if (isspace(c))
		{
			/* Before attribute value */
			break;
		}
		if (c=='>')
		{
			/* End of tag */
			if (context->current_tag->name)
				catch( start_element(context) );
			context->state = S_after_open;
			break;
		}
		else
		if (c=='\'')
		{
			context->state = S_squoted;
			break;
		}
		else
		if (c=='"')
		{
			context->state = S_dquoted;
			break;
		}
		catch( HTChunkPutc(string, c) );
		context->state = S_value;
		break;
	
	case S_value:
		if (isspace(c) || (c=='>'))
		{
			/* End of word */
			catch( HTChunkTerminate(string) );
			catch(handle_attribute_value(context, string->data));
			string->size = 0;
			if (c=='>')
			{
				/* End of tag */
				if (context->current_tag->name)
					catch( start_element(context) );
				context->state = S_after_open;
				break;
			}
			else
				context->state = S_tag_gap;
		}
		else
		{
			catch( HTChunkPutc(string, c) );
		}
		break;
		
	case S_squoted:
		/* Quoted attribute value (using single quotes) */
		if (c=='\'')
		{
			/* End of attribute value */
			catch( HTChunkTerminate(string) );
			catch( handle_attribute_value(context, string->data) );
			string->size = 0;
			context->state = S_tag_gap;
		}
		else
		{
			catch( HTChunkPutc(string, c) );
		}
		break;
	
	case S_dquoted:
		/* Quoted attribute value (using double quotes) */
		if (c=='"')
		{
			/* End of attribute value */
			catch( HTChunkTerminate(string) );
			catch(handle_attribute_value(context, string->data));
			string->size = 0;
			context->state = S_tag_gap;
		}
		else
		{
			catch( HTChunkPutc(string, c) );
		}
		break;
	
	case S_end:
		/* </ */
		if (isalnum(c))
			catch( HTChunkPutc(string, c) );
		else
		{
			/* End of end tag name */
			HTTag * t;
			catch( HTChunkTerminate(string) );

			if (!*string->data)
			{
				/* Empty end tag */
				t = context->element_stack->tag;
			}
			else
			{
				t = SGMLFindTag(dtd, string->data);
			}
			if (t)
			{
				context->current_tag = t;
				end_element( context, context->current_tag);
			}

			string->size = 0;
			context->current_attribute_number = INVALID;
			if (c!='>')
			{
				context->state = S_junk_tag;
			}
			else
			{
				context->state = S_text;
			}
		}
		break;

		
	case S_junk_tag:
		if (c=='>')
		{
			context->state = S_text;
		}
		break;

	}
	return err;
}


Error SGML_string(HTStream *context, const char *s)
{
	Error err=ERR_OK;

	while (*s)
		catch( SGML_character(context, *s++) );
	return err;
}


Error SGML_write(HTStream *context, const char *b, int l)
{
	Error err=ERR_OK;

	while (l-- > 0)
		catch( SGML_character(context, *b++) );
	return err;
}

/*_______________________________________________________________________
*/

/*	Structured Object Class */
/*	----------------------- */

HTStreamClass SGMLParser =
{       
	"SGMLParser",
	SGML_flush,
	SGML_free,
	SGML_abort,
	SGML_character,
	SGML_string,
	SGML_write,
};

/*	Create SGML Engine
**	------------------
**
** On entry,
**	dtd 	represents the DTD, along with
**	actions 	is the sink for the data as a set of routines.
*/

Error SGML_new(SGML_dtd *dtd, HTStructured *target, HTStream **ret)
{
	int i;
	Error err;
	HTStream *context;

	catch( MEMAllocate(sizeof(HTStream),ret) );
	context=*ret;
	catch( HTChunkCreate(128,&context->string) );

	context->isa = &SGMLParser;
	context->dtd = dtd;
	context->target = target;
	context->actions = (HTStructuredClass*)(((HTStream *)target)->isa);
	context->state = S_text;

	return err;
}
